"""
Project Views
"""

from django.http import JsonResponse
from .tools import by_request

def view(request):
    "Example view"
    return JsonResponse(by_request.guess_country_language_currency(request))




"""
Command update Language-Country.
"""

from django.core.management.base import BaseCommand
from ...tools import models
from django_g11n.tools import language_country
# pylint: disable=no-member
SPECIFIERS = dict(language_country.SPECIFIERS)

def update_specifiers():
    "Update Language Country Specifiers"
    model = models.ALL['LanguageCountrySpecifier']
    index_old = set(model.objects.filter(override=False)\
                                               .values_list('short', flat=True))

    data = list(language_country.get()['specifiers'])
    index_new = set([row[0] for row in data])

    db_obsolete = index_old.difference(index_new)
    if len(db_obsolete) > 0:
        model.objects.filter(short__in=db_obsolete).delete()
        print('# %s DB Language Country Specifiers deleted.' % len(db_obsolete))

    db_insert = index_new.difference(index_old)
    tmp = list()
    for entry in data:
        if entry[0] in db_insert:
            instance = model()
            instance.short = entry[0]
            instance.value = entry[1]
            instance.override = False
            tmp.append(instance)

    model.objects.bulk_create(tmp)
    text = '# %s DB Language Country Specifier entries inserted.'
    text = text % len(db_insert)
    print(text)


def lc_string(language, country, specifier):
    "Creates the string."
    tmp = '-'.join([language, country])
    if specifier != '':
        tmp += ' ('+SPECIFIERS[specifier]+')'

    return tmp

def lookups():
    "return lookups"
    tmp = dict()
    tmp['language3'] = dict()
    tmp['language2'] = dict()
    tmp['country'] = dict()
    tmp['specifiers'] = dict()

    for item in models.ALL['Language'].objects.all():
        tmp['language3'][item.bibliographic] = item
        tmp['language2'][item.code_a2] = item

    tmp['country'] = {item.code_2:item for item in \
                                            models.ALL['Country'].objects.all()}
    tmp['specifier'] = {item.short:item for item in \
                           models.ALL['LanguageCountrySpecifier'].objects.all()}
    return tmp

def _create_indexes(data):
    "creates the indexes for update_language_countries."
    index_lup = dict()
    index_old = set()

    for item in models.ALL['LanguageCountry'].objects.filter(override=False):
        string = str(item)
        index_old.add(string)
        index_lup[string] = item

    index_new = set()
    for language, country, specifier, _ in data:
        index_new.add(lc_string(language, country, specifier))

    return index_new, index_old, index_lup

def _find_and_remove_old(index_old, index_new, index_lup):
    "remove old countries"
    db_obsolete = index_old.difference(index_new)

    if len(db_obsolete) > 0:
        delete_ids = list()
        for key in db_obsolete:
            delete_ids.append(index_lup[key].id)
        models.ALL['LanguageCountry'].objects.filter(id__in=delete_ids).delete()
        print('# %s DB Language Countries deleted.' % len(db_obsolete))

def update_language_countries():
    "Update countries"
    data = list(language_country.get()['language_tld'])
    index_new, index_old, index_lup = _create_indexes(data)
    _find_and_remove_old(index_old, index_new, index_lup)

    db_insert = index_new.difference(index_old)
    tmp = list()
    lup = lookups()
    for language, country, specifier, default in data:
        string = lc_string(language, country, specifier)
        if string in db_insert:
            if len(language) > 2:
                language = lup['language3'][language]
            else:
                language = lup['language2'][language]

            country = lup['country'][country]
            if len(specifier) > 0:
                specifier = lup['specifier'][specifier]
            else:
                specifier = None

            instance = models.ALL['LanguageCountry']()
            instance.language_id = language.id
            instance.country_id = country.id
            instance.default = default
            instance.override = False
            if specifier is not None:
                instance.specifier_id = specifier.id

            tmp.append(instance)

    models.ALL['LanguageCountry'].objects.bulk_create(tmp)
    print('# %s DB Language entries inserted.' % len(tmp))

def update_country_names_local():
    "In the model CountryName update language to 'en-US' for empty ones"
    language = models.ALL['LanguageCountry'].objects.get(language__code_a2='en',
                                                         country__code_2='US')
    models.ALL['CountryName'].objects.filter(language__isnull=True)\
                                                      .update(language=language)

def update():
    "Standard function to update data."
    print('#' * 80)
    print('# Update Language Country')
    update_specifiers()
    update_language_countries()
    update_country_names_local()
    print('#' * 80)
    print('')


class Command(BaseCommand):
    """Update Languages"""
    help = "Update language/country combos."

    # pylint: disable=unused-argument
    def handle(self, *args, **kwargs):
        update()




"main testing module"
from django.test import TestCase

# Create your tests here.
class VariousTestCase(TestCase):
    "main test case"
    # pylint: disable=missing-docstring, no-member, invalid-name
    @classmethod
    def setUpClass(cls):
        returns = super(VariousTestCase, cls).setUpClass()
        from . import common
        from ..tools import models
        print('# Setting up IP ranges, this will take a while.')
        common.setup_ipranges_all()
        print('# ... done.')
        common.setup_currencies()
        return returns

    def setUp(self):
        TestCase.setUp(self)
        from django.core.management import call_command
        from . import common
        self.call = common.call_command_returns

    def tearDown(self):
        TestCase.tearDown(self)

    def test_001_guess(self):
        expected = 'GB'
        actually = self.call('guess_tld_by_ip',  '109.74.193.121')
        self.assertEqual(expected, actually)

        provided = 'GB'
        expected = 'GBP'
        actually = self.call('currency_by_tld', provided)
        self.assertEqual(expected, actually)


    def test_002_tools_fetch(self):
        from ..tools import fetch
        self.assertEqual(fetch.country_by_ip('0.0.0.0'), None)
        from ..tools import models
        _ = models.ALL['IPRange'].objects.get(id=1)
        _.id = None
        _.save()
        self.assertRaises(ValueError, fetch.country_by_ip, '3.0.0.1')
        


if __name__ == '__main__': # pragma: no cover
    # pylint: disable=wrong-import-position
    import django
    django.setup()
    django.core.management.call_command('test')




# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-06-28 14:54
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Country',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dts_insert', models.DateTimeField(auto_now_add=True)),
                ('dts_update', models.DateTimeField(blank=True, null=True)),
                ('dts_delete', models.DateTimeField(blank=True, null=True)),
                ('numeric', models.PositiveSmallIntegerField(unique=True)),
                ('code_2', models.CharField(max_length=2, unique=True)),
                ('code_3', models.CharField(max_length=3, unique=True)),
                ('iso3166', models.BooleanField(default=False)),
                ('obsolete', models.BooleanField(default=False)),
            ],
            options={
                'verbose_name_plural': 'Countries',
            },
        ),
        migrations.CreateModel(
            name='CountryName',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dts_insert', models.DateTimeField(auto_now_add=True)),
                ('dts_update', models.DateTimeField(blank=True, null=True)),
                ('dts_delete', models.DateTimeField(blank=True, null=True)),
                ('default', models.BooleanField(default=False)),
                ('value', models.CharField(max_length=64)),
                ('iso3166', models.BooleanField(default=False)),
                ('country', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_g11n.Country')),
            ],
            options={
                'verbose_name_plural': 'Country Names',
            },
        ),
        migrations.CreateModel(
            name='Currency',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dts_insert', models.DateTimeField(auto_now_add=True)),
                ('dts_update', models.DateTimeField(blank=True, null=True)),
                ('dts_delete', models.DateTimeField(blank=True, null=True)),
                ('country', models.TextField()),
                ('numeric', models.PositiveSmallIntegerField()),
                ('name', models.TextField()),
                ('code', models.TextField()),
                ('decimals', models.SmallIntegerField(blank=True, null=True)),
                ('is_fund', models.BooleanField(default=False)),
                ('default', models.BooleanField(default=True)),
                ('iso4217', models.BooleanField(default=False)),
                ('reference', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='django_g11n.Country')),
            ],
            options={
                'verbose_name_plural': 'Currencies',
            },
        ),
        migrations.CreateModel(
            name='IPRange',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dts_insert', models.DateTimeField(auto_now_add=True)),
                ('dts_update', models.DateTimeField(blank=True, null=True)),
                ('dts_delete', models.DateTimeField(blank=True, null=True)),
                ('identifier', models.TextField()),
                ('regional_nic', models.TextField()),
                ('tld', models.CharField(max_length=2)),
                ('ipv', models.PositiveSmallIntegerField()),
                ('network_hex', models.TextField(max_length=32)),
                ('broadcast_hex', models.TextField(max_length=32)),
                ('reference', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='django_g11n.Country')),
            ],
            options={
                'verbose_name_plural': 'IP Ranges',
            },
        ),
        migrations.CreateModel(
            name='Language',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dts_insert', models.DateTimeField(auto_now_add=True)),
                ('dts_update', models.DateTimeField(blank=True, null=True)),
                ('dts_delete', models.DateTimeField(blank=True, null=True)),
                ('bibliographic', models.CharField(blank=True, max_length=7, null=True)),
                ('terminologic', models.CharField(blank=True, max_length=3, null=True)),
                ('code_a2', models.CharField(blank=True, max_length=2, null=True)),
                ('english', models.TextField()),
                ('french', models.TextField()),
                ('iso639_2', models.BooleanField(default=False)),
                ('obsolete', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='LanguageCountry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dts_insert', models.DateTimeField(auto_now_add=True)),
                ('dts_update', models.DateTimeField(blank=True, null=True)),
                ('dts_delete', models.DateTimeField(blank=True, null=True)),
                ('default', models.BooleanField(default=False)),
                ('override', models.BooleanField(default=True)),
                ('country', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_g11n.Country')),
                ('language', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_g11n.Language')),
            ],
            options={
                'verbose_name_plural': 'Language Countries',
            },
        ),
        migrations.CreateModel(
            name='LanguageCountrySpecifier',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dts_insert', models.DateTimeField(auto_now_add=True)),
                ('dts_update', models.DateTimeField(blank=True, null=True)),
                ('dts_delete', models.DateTimeField(blank=True, null=True)),
                ('short', models.CharField(max_length=2)),
                ('value', models.CharField(max_length=64)),
                ('override', models.BooleanField(default=True)),
            ],
            options={
                'verbose_name_plural': 'Language Country Specifiers',
            },
        ),
        migrations.AddField(
            model_name='languagecountry',
            name='specifier',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='django_g11n.LanguageCountrySpecifier'),
        ),
        migrations.AddField(
            model_name='countryname',
            name='language',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='django_g11n.LanguageCountry'),
        ),
    ]



"""
Command update languages.
"""

from django.core.management.base import BaseCommand
from ...tools import models
from django_g11n.tools import language
# pylint: disable=no-member

def update_languages():
    "Update countries"
    model = models.ALL['Language']
    index_old = set(model.objects.all().values_list('bibliographic', flat=True))

    data = list(language.get())
    index_new = set([row[0] for row in data])

    db_obsolete = index_old.difference(index_new)
    if len(db_obsolete) > 0:
        model.objects.filter(bibliographic__in=db_obsolete)\
                                                          .update(obsolete=True)
        print('# %s DB Language entries marked obsolete.' % len(db_obsolete))

    db_insert = index_new.difference(index_old)
    tmp = list()
    for entry in data:
        if entry[0] in db_insert:
            instance = model()
            instance.bibliographic = entry[0]
            instance.terminologic = entry[1]
            instance.code_a2 = entry[2]
            instance.english = entry[3]
            instance.french = entry[4]
            instance.iso639_2 = entry[5]
            tmp.append(instance)

    model.objects.bulk_create(tmp)
    print('# %s DB Language entries inserted.' % len(db_insert))


def update():
    "Standard function to update data."
    print('#' * 80)
    print('# Update Languages')
    update_languages()
    print('#' * 80)
    print('')

class Command(BaseCommand):
    """Update Languages"""
    help = "Update language data."

    # pylint: disable=unused-argument
    def handle(self, *args, **kwargs):
        update()




# Copyright (c) 2016, Martin P. Hellwig, All Rights Reserved.
"""
Test the commands
"""
# pylint:disable=R0904,C0111,W0212, import-error, too-few-public-methods
# pylint:disable=attribute-defined-outside-init, wrong-import-position
from django.test import TestCase
import os
    

class TestCommonIpRanges(TestCase):
    def test_010_get_ip(self):
        from ..tools import ipranges
        from .common import FTPMock
        ipranges.ftplib = FTPMock()
        for _ in ipranges.get():
            pass
        # Running twice, the second one is to test the cache functionality.
        for _ in ipranges.get():
            pass

        for filename in os.listdir(path=ipranges.TWD):
            if 'x_remove_x' in filename:
                path = os.path.join(ipranges.TWD, filename)
                os.remove(path)

    def test_020_parse_row(self):
        tmp = 'arin|US|ipv6|2001:400::|32|19990803|allocated|04f048163e37eef48'\
              'd891498545eefc0'
        from ..tools import ipranges
        self.assertEqual(5, len(ipranges._parse_row(tmp)))

        no_tld = tmp.replace('US', '')
        self.assertEqual(None, ipranges._parse_row(no_tld))

        no_ver = tmp.replace('ipv6', 'asn')
        self.assertEqual(None, ipranges._parse_row(no_ver))

    def test_030_compact_ranges(self):
        from ..tools import ipranges
        compact = ipranges.CompactRanges()
        compact.add(*['A', 'B', 'C', 2, 5])
        compact.add(*['A', 'B', 'C', 6, 9])
        self.assertEqual([['A', 'B', 'C', 2, 9]],
                         compact.ranges)

    def test_040_local_file(self):
        from ..tools import ipranges
        url = 'http://example.com/somewhere/over/the/rainbow'
        self.assertEqual(None, ipranges._local_file_from_url(url))

    def test_050_parse_latest(self):
        from ..tools import ipranges
        url = 'http://example.com/somewhere/over/the/rainbow'
        self.assertEqual(None, ipranges.parse_latest(url))

    def test_050_parse_latest_mock(self):
        import tempfile
        _,name = tempfile.mkstemp()
        os.close(_)

        with open(name, 'w') as file_write:
            file_write.write('\n'*13)

        def mock(_):
            "mock local from url"
            return name

        from ..tools import ipranges
        ipranges._local_file_from_url = mock
        returns = ipranges.parse_latest('')
        os.remove(name)

        self.assertEqual([], returns)


if __name__ == '__main__': # pragma: no cover
    import django
    django.setup()
    django.core.management.call_command('test')



"main testing module"
from django.test import TestCase

class MockUpdate(object):
    def __init__(self, module):
        self.restore = module

    def update(self):
        pass

# Create your tests here.
class UpdateLanguageCountriesTestCase(TestCase):
    "main test case"
    # pylint: disable=missing-docstring, no-member, invalid-name
    @classmethod
    def setUpClass(cls):
        returns = super(UpdateLanguageCountriesTestCase, cls).setUpClass()
        return returns

    def setUp(self):
        TestCase.setUp(self)
        from django.core.management import call_command
        self.call_command = call_command
        #
        from ..management.commands import g11n_setup
        self.g11n_setup = g11n_setup
        self.update_modules = ['update_countries', 'update_currencies',
                               'update_ipranges', 'update_languages',
                               'update_language_countries']
        for name in self.update_modules:
            module = getattr(g11n_setup, name)
            mock = MockUpdate(module)
            setattr(g11n_setup, name, mock)

    def tearDown(self):
        for name in self.update_modules:
            mock = getattr(self.g11n_setup, name)
            setattr(self.g11n_setup, name, mock.restore)

        TestCase.tearDown(self)

    def test_001_insert(self):
        self.call_command('g11n_setup')


if __name__ == '__main__': # pragma: no cover
    # pylint: disable=wrong-import-position
    import django
    django.setup()
    django.core.management.call_command('test')




"""
Fetch Currency data, ISO 4217
"""
from io import StringIO
from xml.etree import ElementTree
import requests

URL = "http://www.currency-iso.org/dam/downloads/lists/list_one.xml"

ADDITIONS = [
  ['Palestine, State of',
   '376', 'New Israeli Sheqel', 'ILS', '2', False, True, False],

  ['South Georgia and the South Sandwich Islands',
   '826', 'Pound Sterling', 'GBP', '2', False, True, False]]

DEFAULTS_OVERRIDE = {'BHUTAN':'BTN',
                     'BOLIVIA (PLURINATIONAL STATE OF)':'BOB',
                     'CHILE':'CLP',
                     'COLOMBIA':'COP',
                     'CUBA':'CUP',
                     'EL SALVADOR':'USD',
                     'HAITI':'HTG',
                     'LESOTHO':'LSL',
                     'MEXICO':'MXN',
                     'NAMIBIA':'NAD',
                     'PANAMA':'USD',
                     'SWITZERLAND':'CHF',
                     'UNITED STATES OF AMERICA (THE)':'USD',
                     'URUGUAY':'UYU'}

def get(url=URL):
    """Fetch currency and parse it.
    Yields per yield:
    - iso_short_name_lower_case
    - currency_numeric
    - currency_name
    - currency_cod,
    - is_fund
    - default
    - boolean if this is an ISO 4217 entry (False if it is an local override)
    All fields are text fields, the iso_short_name_lower_case is all upper case
    due to the source document.
    """
    response = requests.get(url)
    response.encoding = 'UTF-8'
    xml = StringIO()
    xml.write(response.text)

    xml.seek(0)
    tree = ElementTree.parse(xml)

    for entry in tree.getiterator():
        if entry.tag != 'CcyNtry':
            continue

        items = ['CtryNm', 'CcyNbr', 'CcyNm', 'Ccy', 'CcyMnrUnts']
        tmp = list()
        empty = False
        is_fund = False
        for item in items:
            element = entry.find(item)
            if element is None:
                empty = True
                element = ''
            else:
                if item == 'CcyNm':
                    if element.get('IsFund') is not None:
                        is_fund=True
                data = element.text.strip()
                data = data.replace('’', "'")
                data = data.replace('’', "'")

            tmp.append(data)
        tmp.append(is_fund)
        tmp.append(True)

        if tmp[0] in DEFAULTS_OVERRIDE:
            if tmp[3] != DEFAULTS_OVERRIDE[tmp[0]]:
                tmp[-1] = False

        # Make ZZ type currencies non-default
        if tmp[0][:2] == 'ZZ' and tmp[0][2:4].isdigit():
            tmp[-1] = False

        if not empty:
            tmp.append(True)
            yield tmp

        for item in ADDITIONS:
            yield item




"""
These are the models at the core of the functionality.
"""
from django.db import models
from django.utils.timezone import now
# pylint: disable=too-few-public-methods, no-member

class QueryManager(models.Manager):
    "Query Manager to restrict returning data"
    def get_queryset(self):
        "get_queryset override"
        query = models.Manager.get_queryset(self)
        query.filter(dts_delete__isnull=False)
        return query

class Abstract(models.Model):
    "Abstract share by all subsequent models."
    objects = QueryManager()

    class Meta:
        "Meta section to identify this is abstract"
        abstract = True

    dts_insert = models.DateTimeField(auto_now_add=True)
    dts_update = models.DateTimeField(null=True, blank=True)
    dts_delete = models.DateTimeField(null=True, blank=True)

    def save(self, force_insert=False, force_update=False, using=None,
        update_fields=None):
        if self.id not in [0, None]:
            self.dts_update = now()
        returns = super().save(force_insert=force_insert,
                               force_update=force_update,
                               using=using, update_fields=update_fields)
        return returns


class Language(Abstract):
    "Stores Languages."
    bibliographic = models.CharField(max_length=7, null=True, blank=True)
    terminologic = models.CharField(max_length=3, null=True, blank=True)
    code_a2 = models.CharField(max_length=2, null=True, blank=True)
    english = models.TextField()
    french = models.TextField()
    iso639_2 = models.BooleanField(default=False)
    obsolete = models.BooleanField(default=False)


    def __str__(self):
        _ = [self.bibliographic, self.terminologic, self.code_a2, self.english,
             self.french]
        _ = [item if item is not None else '' for item in _ ]
        return ' | '.join(_)


class Country(Abstract):
    "Country Code Top Level domains, includes regions."
    class Meta:
        "Set verbose name"
        verbose_name_plural = 'Countries'

    numeric = models.PositiveSmallIntegerField(unique=True)
    code_2 = models.CharField(max_length=2, unique=True)
    code_3 = models.CharField(max_length=3, unique=True)
    iso3166 = models.BooleanField(default=False)
    obsolete = models.BooleanField(default=False)

    def __str__(self):
        return self.code_2


class LanguageCountrySpecifier(Abstract):
    "Language Country specifier."
    short = models.CharField(max_length=2)
    value = models.CharField(max_length=64)
    override = models.BooleanField(default=True)

    class Meta:
        "Set verbose name"
        verbose_name_plural = 'Language Country Specifiers'

    def __str__(self):
        return self.value


class LanguageCountry(Abstract):
    "Language Country codes."
    language = models.ForeignKey(Language)
    country = models.ForeignKey(Country)
    specifier = models.ForeignKey(LanguageCountrySpecifier,
                                  null=True, blank=True)
    default = models.BooleanField(default=False)
    override = models.BooleanField(default=True)

    class Meta:
        "Set verbose name"
        verbose_name_plural = 'Language Countries'

    def __str__(self):
        if self.language.code_a2 is None:
            _ = [str(self.language.bibliographic)]
        else:
            _ = [str(self.language.code_a2)]

        _.append(self.country.code_2)

        tmp = '-'.join(_)
        if self.specifier is not None:
            tmp += ' ('+self.specifier.value+')'

        return tmp


class CountryName(Abstract):
    "Name of the country."
    class Meta:
        "Set verbose name"
        verbose_name_plural = 'Country Names'

    country = models.ForeignKey(Country)
    default = models.BooleanField(default=False)
    language = models.ForeignKey(LanguageCountry, null=True, blank=True)
    value = models.CharField(max_length=64)
    iso3166 = models.BooleanField(default=False)

    def __str__(self):
        return self.value


class Currency(Abstract):
    "Stores country and currency"
    class Meta:
        "Set verbose name"
        verbose_name_plural = 'Currencies'

    country = models.TextField()
    reference = models.ForeignKey(Country, null=True, blank=True)
    numeric = models.PositiveSmallIntegerField()
    name = models.TextField()
    code = models.TextField()
    decimals = models.SmallIntegerField(null=True, blank=True)
    is_fund = models.BooleanField(default=False)
    default = models.BooleanField(default=True)
    iso4217 = models.BooleanField(default=False)

    def __str__(self):
        return self.code + ' ' + str(self.numeric)



class IPRange(Abstract):
    "Stores IP ranges for each Country or Region TLD"
    class Meta:
        "Set verbose name"
        verbose_name_plural = 'IP Ranges'

    identifier = models.TextField()
    regional_nic = models.TextField()
    tld = models.CharField(max_length=2)
    reference = models.ForeignKey(Country, null=True, blank=True)
    ipv = models.PositiveSmallIntegerField()
    network_hex = models.TextField(max_length=32)
    broadcast_hex = models.TextField(max_length=32)

    def __str__(self):
        return self.identifier





# Copyright (c) 2016, Martin P. Hellwig, All Rights Reserved.
"""
Test the commands
"""
# pylint:disable=R0904,C0111,W0212, import-error, too-few-public-methods
# pylint:disable=attribute-defined-outside-init, wrong-import-position
# pylint:disable=no-self-use
from django.test import TestCase

ORDER = ['Language', 'Country', 'LanguageCountrySpecifier', 'LanguageCountry',
         'CountryName', 'Currency', 'IPRange']
KWARG = {'Language':{'english':'Oi', 'french':'Ey'},
         'Country':{'numeric':999, 'code_2':'XX', 'code_3':'XXX'},
         'LanguageCountrySpecifier':{'short':'XX', 'value':'X all the way'},
         'LanguageCountry':{'language':'Language', 'country':'Country',},
         'CountryName':{'country':'Country', 'value':'X'},
         'Currency':{'country':'X', 'numeric':1, 'name':'X', 'code':'X'},
         'IPRange':{'identifier':'X', 'regional_nic':'X', 'tld':'XX',
                    'ipv':4, 'network_hex':'X', 'broadcast_hex':'X'},
         }

class TestModelsString(TestCase):
    def test_010_string(self):
        from ..tools import models
        tmp = dict()
        for key in ORDER:
            kwarg = KWARG[key]
            model = models.ALL[key]
            for kwarg_key in list(kwarg):
                kwarg_value = kwarg[kwarg_key]
                if kwarg_value in tmp:
                    kwarg[kwarg_key] = tmp[kwarg_value]

            tmp[key] = model.objects.create(**kwarg)
            str(tmp[key])

    def test_020_resave(self):
        from ..tools import models
        key = 'Language'
        model = models.ALL[key]
        subject = model.objects.create(**KWARG[key])
        subject.code_a2 = 'en'
        subject.save()

if __name__ == '__main__': # pragma: no cover
    import django
    django.setup()
    django.core.management.call_command('test')



"""
Get Country data, we use the django_countries app to fetch the data.
The app just uses a manually fetched data from the ISO site, since they already
have that coded down I am not repeating that here, even though I have to go to
some troubles recombining it.
"""
from collections import defaultdict
from django_countries import data

ADDITIONS = [[9995, 'CB', 'CBN', False, 'Caribbean'],
             [9996, 'AP', 'APA', False, 'Asia and Pacific'],
             [9997, 'EU', 'EUR', False, 'Europe', 'European Union'],
             [9998, 'ZZ', 'ZZR', False, 'IETF Reserved'],
             [9999, '??', '???', False, 'Unknown']]

def get():
    """Reconstruct the ISO 3166 data, returns a list of row where each row is:
    - Country numeric code
    - Two letter code
    - Three letter code
    - Boolean if it is ISO 3166 (False if it is a local override.
    - Default label of the country
    - Additional label of the country
    Note that there may be multiple additional labels at the end.
    """
    tmp = defaultdict(list)
    for collection in data.COUNTRIES, data.COMMON_NAMES:
        for entry in collection.items():
            key = entry[0]
            value =  entry[1].__dict__['_proxy____args'][0]
            tmp[key].append(value)

    returns = list()
    for key in tmp:
        _ = [data.ALT_CODES[key][1], key, data.ALT_CODES[key][0]]
        _.append(True)
        _ += tmp[key][::-1] # Reverse the Country labels

        returns.append(_)
    returns.sort()
    returns += ADDITIONS
    return returns

    


"""
Project information.
We use this module as __init__ is a generic module that django requires to be
specifically set up.
"""
LABELS = dict()
LABELS['class'] = 'G11N'
LABELS['name'] = 'django_g11n'
LABELS['verbose_name'] = 'Globalisation'



"""
Command update IP Ranges.
"""

from django.core.management.base import BaseCommand
from ...tools import models
from django_g11n.tools import ipranges
# pylint: disable=no-member

def fetch_and_insert():
    "Fetch the data from the regional NIC."
    model = models.ALL['IPRange']
    index_old = set(model.objects.all().values_list('identifier',
                                                    flat=True))

    data = list(ipranges.get())
    index_new = set([row[0] for row in data])

    db_remove = index_old.difference(index_new)
    model.objects.filter(identifier__in=db_remove).delete()
    print('# %s entries removed from DB' % len(db_remove))

    db_insert = index_new.difference(index_old)

    lookup = dict(\
                models.ALL['Country'].objects.all().values_list('code_2', 'id'))

    tmp = list()
    not_found = set()
    for entry in data:
        if entry[0] in db_insert:
            instance = model()
            instance.identifier = entry[0]
            instance.regional_nic = entry[1]
            instance.tld = entry[2]
            if entry[2] in lookup:
                instance.reference_id = lookup[entry[2]]
            else:
                not_found.add(entry[2])

            instance.ipv = entry[3]
            instance.network_hex = entry[4]
            instance.broadcast_hex = entry[5]
            tmp.append(instance)

    if len(not_found) > 0:
        print("! Couldn't find country/region entries for: %s" % not_found)
    print("# Bulk insert %s items into the DB ..." % len(db_insert))
    model.objects.bulk_create(tmp)
    print('# Bulk insert done.')


def update():
    "Standard function to update data."
    fetch_and_insert()


class Command(BaseCommand):
    """Update IP Ranges"""
    help = "Use ipranges module to update the IP Ranges from the regional NIC."

    # pylint: disable=unused-argument
    def handle(self, *args, **kwargs):
        update()



"""
Command Guess TLD by IP
"""

from django.core.management.base import BaseCommand
from ...tools import fetch
# pylint: disable=no-member

_ARGUMENT_ADDRESS = 'IP Address'
class Command(BaseCommand):
    """Guess the TLD by given IP."""
    help = "Guess the TLD by given IP."

    def add_arguments(self, parser):
        parser.add_argument(_ARGUMENT_ADDRESS)

    # pylint: disable=unused-argument
    def handle(self, *args, **kwargs):
        address = kwargs[_ARGUMENT_ADDRESS]
        fetched = fetch.country_by_ip(address)
        return fetched



'''
Created on 23 Jun 2016

@author: martin
'''
import json
import zipfile
import os
from io import StringIO
import requests

_STATE = {'mlsd':False}
# pylint: disable=missing-docstring, no-self-use, unused-argument
# pylint: disable=protected-access, too-few-public-methods


def make_data_path(name):
    "Make the data path."
    _ = os.path.dirname(os.path.abspath(__file__))
    _ = os.path.join(_, 'data', name)
    return _


def bulk_json_insert(data):
    "Bulk insert json"
    tmp = dict()
    data = json.loads(data)
    from django.apps import apps
    from ..tools import models
    for item in data:
        model_name = item['model']
        model = apps.get_model(*model_name.rsplit('.', 1))
        fields = item['fields']
        fields['id'] = item['pk']

        for key in list(fields):
            field = model._meta.get_field(key)
            if field.is_relation:
                fields[key+'_id'] = fields.pop(key)

        if model not in tmp:
            tmp[model] = list()

        tmp[model].append(model(**fields))

    for model in list(tmp.keys()):
        model.objects.bulk_create(tmp[model])


def setup_ipranges_all():
    "Setup all ip's."
    from ..tools import models
    path = make_data_path('country.json')
    with open(path, 'r') as file_open:
        data = ''.join(file_open.readlines())
        bulk_json_insert(data)

    name = 'iprange_2016_06_23.json.zip'
    path = make_data_path(name)
    unzip = zipfile.ZipFile(path)
    data = unzip.read(name.rsplit('.', 1)[0]).decode('UTF-8')
    bulk_json_insert(data)


def call_command_returns(*args, **kwargs):
    "call command but wich returns the output."
    from django.core.management import call_command
    stdout = StringIO()
    kwargs['stdout'] = stdout
    call_command(*args, **kwargs)
    return stdout.getvalue().strip()


class RequestsMock(object):
    "Mock the request module"
    def __init__(self):
        self._restores = dict()
        self._response = dict()
        self.text = None

    def add_response_text_from_data(self, url, file_name):
        "If url is called the content of file_name is returned"
        _ = make_data_path(file_name)
        with open(_, 'r') as file_open:
            self._response[url] = ''.join(file_open.readlines())

    def insert_mock(self):
        "Insert the mock"
        self._restores['get'] = requests.get
        requests.get = self.get

    def remove_mock(self):
        "Remove the mock"
        requests.get = self._restores['get']

    def get(self, url):
        "Mocked request get function"
        self.text = self._response[url]
        return self


def setup_currencies():
    "Insert all currencies"
    from django.core.management import call_command
    call_command('update_countries')

    mock = RequestsMock()
    url = 'http://www.currency-iso.org/dam/downloads/lists/list_one.xml'
    file_name = 'list_one.xml'
    mock.add_response_text_from_data(url, file_name)
    mock.insert_mock()
    mock.remove_mock()
    call_command('update_currencies')


class MockFTP():
    def __init__(self, host):
        self._host = host
        self._path = None
        self._mlsd_called = False

        tmp = {'size':'91',
               'modify':'x_remove_x'}
        self._files = [
            ['delegated-arin-extended-latest', tmp],
            ['delegated-ripencc-extended-latest', tmp],
            ['delegated-afrinic-extended-latest', tmp],
            ['delegated-apnic-extended-latest', tmp],
            ['delegated-lacnic-extended-latest', tmp],]

    def login(self):
        pass

    def cwd(self, path):
        self._path = path

    def mlsd(self):
        if _STATE['mlsd']:
            _STATE['mlsd'] = False
            raise ValueError()
        else:
            _STATE['mlsd'] = True

        return self._files

    def retrbinary(self, retr, callback):
        ip4 = 'arin|US|ipv4|12.0.0.0|16777216|19830823|allocated|99e9610432f009e0e177ba0c274bb288\n'
        ip6 = 'arin|US|ipv6|2001:418::|32|20000524|allocated|9f14454567a6c23e60bfd4fec24d1438\n'
        nic = 'arin||ipv4|13.0.0.0|16777216|19830823|allocated|99e9610432f009e0e177ba0c274bb288\n'
        asn = 'arin|US|asn|14.0.0.0|16777216|19830823|allocated|99e9610432f009e0e177ba0c274bb288\n'
        for entry in [ip4, ip6, nic, asn]:
            callback(entry.encode('ASCII'))

    def retrlines(self, command, callback):
        if not 'second' in command:
            callback('-> second')
        else:
            callback('one two three four five six seven eight')



class FTPMock(object):
    FTP = MockFTP
    error_perm = ValueError





"main testing module"
from django.test import TestCase

# Create your tests here.
class UpdateCountriesTestCase(TestCase):
    "main test case"
    # pylint: disable=missing-docstring, no-member, invalid-name
    @classmethod
    def setUpClass(cls):
        returns = super(UpdateCountriesTestCase, cls).setUpClass()
        return returns

    def setUp(self):
        TestCase.setUp(self)

    def test_001_insert(self):
        from django.core.management import call_command
        call_command('update_countries')

        from ..tools import models

        model = models.ALL['Country']
        query = model.objects.filter()
        query = query.values_list('code_2', flat=True).distinct()
        self.assertEqual(254, query.count())

        model = models.ALL['CountryName']
        query = model.objects.filter()
        query = query.values_list('value', flat=True).distinct()
        self.assertEqual(270, query.count())



if __name__ == '__main__': # pragma: no cover
    # pylint: disable=wrong-import-position
    import django
    django.setup()
    django.core.management.call_command('test')




"main testing module"
from django.test import TestCase

# Create your tests here.
class UpdateLanguagesTestCase(TestCase):
    "main test case"
    # pylint: disable=missing-docstring, no-member, invalid-name
    @classmethod
    def setUpClass(cls):
        returns = super(UpdateLanguagesTestCase, cls).setUpClass()
        return returns

    def setUp(self):
        returns = TestCase.setUp(self)
        from django.core.management import call_command
        self.call_command = call_command

        from . import common
        self.mock = common.RequestsMock()
        url = 'http://www.loc.gov/standards/iso639-2/ISO-639-2_utf-8.txt'
        file_name = 'ISO-639-2_utf-8.txt'

        self.mock.add_response_text_from_data(url, file_name)
        self.mock.insert_mock()

        return returns

    def tearDown(self):
        TestCase.tearDown(self)
        self.mock.remove_mock()


    def test_001_update_language(self):
        self.call_command('update_languages')
        # Calling twice, as the second time it should exclude it.
        #pylint: disable=protected-access, undefined-variable
        #
        from ..tools import models
        models.ALL['Language'].objects.create(
            code_a2='**', english='Double Star', french='Astérisque Double')
        self.call_command('update_languages')

    def test_002_update_language_country(self):
        self.call_command('update_countries')
        self.call_command('update_languages')
        self.call_command('update_language_countries')
        #
        from ..tools import models
        models.ALL['LanguageCountrySpecifier'].objects.create(
            short='??', value='Question Marks', override=False)
        model = models.ALL['LanguageCountry']()
        model.language_id=4
        model.country_id=16
        model.specifier_id=1
        model.override=False
        model.save()
        self.call_command('update_language_countries')

    def test_003_tool_country_addition(self):
        from ..tools import language
        addition = [1,2,3,4,5,6]
        language.ADDITIONS.append(addition)
        is_added = False
        for entry in language.get():
            if entry == addition:
                is_added = True
                break

        self.assertTrue(is_added)
        


if __name__ == '__main__': # pragma: no cover
    # pylint: disable=wrong-import-position
    import django
    django.setup()
    django.core.management.call_command('test')




"""
Command update country.
"""

from django.core.management.base import BaseCommand
from ...tools import models, country
# pylint: disable=no-member


def update_countries():
    "Update countries"
    model = models.ALL['Country']
    index_old = set(model.objects.all().values_list('numeric', flat=True))

    data = list(country.get())
    index_new = set([row[0] for row in data])

    db_obsolete = index_old.difference(index_new)
    model.objects.filter(numeric__in=db_obsolete).update(obsolete=True)
    print('# %s DB Country entries marked obsolete.' % len(db_obsolete))

    db_insert = index_new.difference(index_old)
    tmp = list()
    for entry in data:
        if entry[0] in db_insert:
            instance = model()
            instance.numeric = entry[0]
            instance.code_2 = entry[1]
            instance.code_3 = entry[2]
            instance.iso3166 = entry[3]
            tmp.append(instance)

    model.objects.bulk_create(tmp)
    print('# %s DB Country entries inserted.' % len(db_insert))

def update_country_names():
    "Update country names"
    tmp = list()
    lookup = dict()
    for item in models.ALL['Country'].objects.all():
        lookup[item.numeric] = item

    model = models.ALL['CountryName']
    index_old = set(model.objects.all().values_list('value', flat=True))

    data = list(country.get())
    for entry in data:
        default=True
        for name in entry[4:]:
            if name not in index_old:
                instance = model()
                instance.country = lookup[entry[0]]
                instance.default = default
                instance.value = name
                instance.iso3166 = entry[3]
                tmp.append(instance)
                if default:
                    default=False
    model.objects.bulk_create(tmp)
    print('# %s DB Country Name entries inserted.' % len(tmp))


def update():
    "Standard function to update data."
    print('#' * 80)
    print('# Update Countries')
    update_countries()
    update_country_names()
    print('#' * 80)
    print('')

class Command(BaseCommand):
    """Update Countries"""
    help = "Update Country data."

    # pylint: disable=unused-argument
    def handle(self, *args, **kwargs):
        update()




"""Command Initial Inserts"""

from django.core.management.base import BaseCommand
from . import (update_countries, update_currencies, update_ipranges,
               update_languages, update_language_countries)
# pylint: disable=no-member

class Command(BaseCommand):
    """Initial setup of the data."""
    help = "Insert initial data for setup purpose"

    # pylint: disable=unused-argument
    def handle(self, *args, **kwargs):
        # For first setup, we need to call the updates in a specific order.
        # Later updates can be done in any order.
        update_countries.update()
        update_currencies.update()
        update_ipranges.update()
        update_languages.update()
        update_language_countries.update()



"main testing module"
from django.test import TestCase

# pylint: disable=too-few-public-methods
class MockToolsIpranges(object):
    "Mock the currency module"
    def __init__(self):
        # identifier, regional_nic, tld, ipv, network_hex, broadcast_hex

        self._data = [
            ['1', 'NIC', 'GB', '4', '00000000', 'ffffffff'],
            ['2', 'NOC', '--', '4', '00000000', 'ffffffff'],
        ]

    def get(self):
        "Mock the get function"
        return self._data

# Create your tests here.
class UpdateIPRangesTestCase(TestCase):
    "main test case"
    # pylint: disable=missing-docstring, no-member, invalid-name
    @classmethod
    def setUpClass(cls):
        returns = super(UpdateIPRangesTestCase, cls).setUpClass()
        return returns

    def setUp(self):
        TestCase.setUp(self)
        from django.core.management import call_command
        self._call_command = call_command
        self._call_command('update_countries')

        from ..management.commands import update_ipranges
        self._update = update_ipranges
        self._restore = update_ipranges.ipranges

        update_ipranges.ipranges = MockToolsIpranges()

    def tearDown(self):
        TestCase.tearDown(self)
        self._update.ipranges = self._restore

    def test_001_insert(self):
        self._call_command('update_ipranges')
        # Calling twice, as the second time it should exclude it.
        self._call_command('update_ipranges')



if __name__ == '__main__': # pragma: no cover
    # pylint: disable=wrong-import-position
    import django
    django.setup()
    django.core.management.call_command('test')




"""
Command Currency by TLD.
"""

from django.core.management.base import BaseCommand
from ...tools import fetch
# pylint: disable=no-member

_ARGUMENT_TLD = 'Country Code Top Level Domain'
class Command(BaseCommand):
    """Return the currency by given top level domain."""
    help = "Return the currency by given top level domain."

    def add_arguments(self, parser):
        parser.add_argument(_ARGUMENT_TLD)

    # pylint: disable=unused-argument
    def handle(self, *args, **kwargs):
        tld = kwargs[_ARGUMENT_TLD]
        fetched = fetch.currency_by_country_tld(tld)
        if fetched != None:
            return str(fetched)



"main testing module"
from django.test import TestCase, Client

# Create your tests here.
class VariousTestCase(TestCase):
    "main test case"
    # pylint: disable=missing-docstring, no-member, invalid-name, no-self-use
    def test_001(self):
        client = Client()
        client.get('', REMOTE_ADDR='109.74.193.121')

    def test_002_by_request_ipaddress(self):
        from ..tools import by_request
        import os
        os.environ['G11N_IP_OVERRIDE'] = ''
        self.assertRaises(ValueError, by_request.ipaddress, None)

    def test_003_by_request_ipaddress(self):
        from ..tools import by_request
        import os
        address = '109.74.193.121'
        os.environ['G11N_IP_OVERRIDE'] = address
        expected = {'ip':address}
        self.assertEqual(expected, by_request.ipaddress(None))

if __name__ == '__main__': # pragma: no cover
    # pylint: disable=wrong-import-position
    import django
    django.setup()
    django.core.management.call_command('test')




"""
Command update currencies.
"""

from django.core.management.base import BaseCommand
from ...tools import models
from django_g11n.tools import currency
# pylint: disable=no-member

def format_search_key(key):
    "Currency names are unfortunately not all directly mappable."
    key = key.lower()
    key = key.replace('(the)', '')
    key = key.strip()
    return key

def insert_delete():
    "Delete old entries insert new"
    model = models.ALL['Currency']
    tmp = list()
    lookup = {key.strip().lower(): value for (key, value) in \
              models.ALL['CountryName'].objects.all()\
                                           .values_list('value', 'country__id')}

    existing = list(model.objects.all().values_list('numeric', flat=True))

    for entry in currency.get():
        if int(entry[1]) in existing:
            continue

        instance = model()
        instance.country = entry[0]
        search = format_search_key(entry[0])

        if search in lookup:
            instance.reference_id = lookup[search]
        else:
            instance.reference_id = lookup['unknown']

        instance.numeric = entry[1]
        instance.name = entry[2]
        instance.code = entry[3]
        if entry[4].isdigit():
            instance.decimals = entry[4]
        instance.is_fund = entry[5]
        instance.default = entry[6]
        instance.iso4217 = entry[7]
        tmp.append(instance)

    if len(tmp) > 0:
        model.objects.all().delete()
        model.objects.bulk_create(tmp)

    print('# %s entries inserted into DB' % len(tmp))

def update():
    "Standard function to update data."
    print('#' * 80)
    print('# Update Currencies')
    insert_delete()
    print('#' * 80)
    print('')

class Command(BaseCommand):
    """Update Currencies"""
    help = "Update all currencies."

    # pylint: disable=unused-argument
    def handle(self, *args, **kwargs):
        update()




"main testing module"
from django.test import TestCase

# Create your tests here.
class UpdateCurrenciesTestCase(TestCase):
    "main test case"
    # pylint: disable=missing-docstring, no-member, invalid-name
    @classmethod
    def setUpClass(cls):
        returns = super(UpdateCurrenciesTestCase, cls).setUpClass()
        return returns

    def setUp(self):
        returns = TestCase.setUp(self)
        from . import common
        common.setup_currencies()
        return returns

    def test_001_insert(self):
        from django.core.management import call_command
        call_command('update_currencies')
        # Calling twice, as the second time it should exclude it.
        call_command('update_currencies')



if __name__ == '__main__': # pragma: no cover
    # pylint: disable=wrong-import-position
    import django
    django.setup()
    django.core.management.call_command('test')




"""
We call all this guessing as there are no certainties.
The client could be using a proxy or vpn or there may be something going on at
the hosting provider that influences the result.

So data from here should be treated as a default that the user can override
where necessary.
"""
import os
from functools import lru_cache
import ipware.ip
from . import fetch

@lru_cache(maxsize=8)
def ipaddress(request=None):
    """Return an IP Address from the request,
    or G11N_IP_OVERRIDE environment variable if that is set."""
    override = None
    key = 'G11N_IP_OVERRIDE'
    if key in os.environ:
        override = os.environ[key]

    if (override is None or override.strip() == '') and request is None:
        text = "Environment variable %s must be set to an appropriate IP."
        raise ValueError(text)

    if request is None:
        value = override
    else:
        value = ipware.ip.get_ip(request)

    return {'ip':value}

def guess_country(request):
    "Guess the country from the request (using the IP address)."
    ip_address = ipaddress(request)['ip']
    return {'country':fetch.country_by_ip(ip_address)}

def guess_language(request):
    "Guess the language from the request (using the browser headers)."
    return {'language':request.LANGUAGE_CODE}

def guess_currency(request):
    "Guess the currency from the request (via guess_country)."
    tld = guess_country(request)['country']
    return {'currency':fetch.currency_by_country_tld(tld)}

def guess_country_language_currency(request):
    "Guess all of these in a single function."
    functions = [guess_country, guess_language, guess_currency, ipaddress]
    tmp = dict()
    for function in functions:
        for key, value in function(request).items():
            tmp[key] = value
    return tmp



"""
PyPi Setup file.
"""
# pylint: disable=no-name-in-module, import-error
from setuptools import setup, find_packages
VERSION = '1.0.2.3'
BASE_URL = "https://bitbucket.org/hellwig/django-g11n"

setup(
  name = 'Django-G11N',
  packages = find_packages(),
  version = VERSION,
  description = 'Django Globalisation Tools',
  author = 'Martin P. Hellwig',
  author_email = 'martin.hellwig@gmail.com',
  url = BASE_URL,
  download_url = BASE_URL + '/get/' + VERSION + '.zip',
  keywords = ['django'],
  license = 'BSD',
  classifiers = ['Programming Language :: Python :: 3',],
  install_requires = ['Django>=1.9.6',
                      'django-countries>=3.4.1',
                      'django-ipware>=1.1.5',
                      'django-integrator>=0.1.0',
                      'requests>=2.10.0'],
)



'''
Created on 9 May 2016

@author: martin
'''
import requests

URL = 'http://www.loc.gov/standards/iso639-2/ISO-639-2_utf-8.txt'
# From the source, the following description:
# These files may be used to download the list of language codes with their
# language names, for example into a database. To read the files, please note
# that one line of text contains one entry. An alpha-3 (bibliographic) code,
# an alpha-3 (terminologic) code (when given), an alpha-2 code (when given),
# an English name, and a French name of a language are all separated by pipe (|)
# characters. If one of these elements is not applicable to the entry, the field
# is left empty, i.e., a pipe (|) character immediately follows the preceding
# entry. The Line terminator is the LF character.
ADDITIONS = []

def get(url=URL):
    """Fetch Languages and parse it, yields an array with:
    - bibliographic code
    - terminologic code
    - alpha-2 code
    - Name in English
    - Name in French.
    - ISO-639-2 boolean (False if it is a local addition)
    """
    response = requests.get(url)
    response.encoding = 'UTF-8'
    text = response.text[::]
    if text.startswith('\ufeff'):
        text = text[1:]

    for line in text.split('\n'):
        row = line.split('|')
        row.append(True)
        yield row

    for row in ADDITIONS:
        yield row

    


"""
Module to fetch and parse regional NIC delegation data
"""
import urllib.parse
import ftplib
import os
from functools import lru_cache
import socket
import ipaddress
from binascii import hexlify
import tempfile

TWD = tempfile.gettempdir()

DELEGATES = [
    # America (non-latin)
    "ftp://ftp.arin.net/pub/stats/arin/delegated-arin-extended-latest",
    # Europe
    "ftp://ftp.ripe.net/ripe/stats/delegated-ripencc-extended-latest",
    # Africa
    "ftp://ftp.afrinic.net/pub/stats/afrinic/delegated-afrinic-extended-latest",
     # Asia & Pacific
    "ftp://ftp.apnic.net/pub/stats/apnic/delegated-apnic-extended-latest",
    # Latin-America
    "ftp://ftp.lacnic.net/pub/stats/lacnic/delegated-lacnic-extended-latest",]


@lru_cache(None)
def _split_url(url):
    "Split delegate url into host, file_path and file_name."
    url = urllib.parse.urlparse(url)
    host = url.netloc
    file_path, file_name = url.path.rsplit('/', 1)
    return (host, file_path, file_name)

def _file_details(ftp, file_name):
    "Retrieve details of the file."
    details = None
    print('# Retrieving file details')
    try:
        listing = list(ftp.mlsd())
        print('# Server support mlsd, extracting details ...')
        for entry in listing:
            name, facts = entry
            if name.lower() == file_name.lower():
                details = facts
                details['name_local'] = name
                details['name_remote'] = name
                break

    except ftplib.error_perm:
        print('# Server does not support mlsd, falling back.')
        tmp = list()
        ftp.retrlines('LIST %s' % file_name, callback=tmp.append)
        if '->' in tmp[0]:
            print('# Fall back: entry is a symbolic link, following ...')
            link2name = tmp[0].split('->')[1].strip()
            tmp = list()
            ftp.retrlines('LIST %s' % link2name, callback=tmp.append)

        details = dict()
        tmp = tmp[0]
        tmp = tmp.rsplit(' ', 1)[0]
        details['name_local'] = file_name
        details['name_remote'] = link2name
        tmp, details['size'], month, day, time = tmp.rsplit(' ', 4)
        details['modify'] = '_'.join([month, day, time.replace(':', '')])

    return details

def download(url):
    "Download the url."
    host, file_path, file_name = _split_url(url)

    print('# Connecting to: %s' % host)
    ftp = ftplib.FTP(host)

    print('# Logging in ...')
    ftp.login()

    print('# Changing cwd to: %s' % file_path)
    ftp.cwd(file_path)

    details = _file_details(ftp, file_name)

    file_cache = '_'.join([details['name_local'],
                           details['size'],
                           details['modify']])
    file_cache += '.csv'

    if file_cache in os.listdir(TWD):
        print('# File is already downloaded !')
        return

    print('# Downloading ...')

    retr = 'RETR %s' % details['name_remote']
    local_file = os.path.join(TWD, file_cache)
    ftp.retrbinary(retr, open(local_file, 'wb').write)

    print('# Downloaded!')


# The parsing part of the program
def _address_range_ipv4(address, width):
    "Convert IPv4 address and amount to integer range."
    # The width of ipv4 addresses is given in number of addresses which
    # are not bounded by exact netmasks for example a width of 640 addresses.
    blocks = address.split('.')
    for index, block in enumerate(blocks):
        blocks[index] = bin(int(block, 10))[2::].zfill(8)
    blocks = ''.join(blocks)
    network = int(blocks, 2)
    broadcast = network + int(width) - 1
    return(network, broadcast)

def _ipv6_to_int(ipv6_address):
    "Convert an IPv6 address to an integer"
    packed_string = socket.inet_pton(socket.AF_INET6, ipv6_address.exploded)
    return int(hexlify(packed_string), 16)

def _address_range_ipv6(address, width):
    "Convert IPv6 address and broadcast to integer range."
    network = ipaddress.ip_network(address+'/'+width)
    broadcast = _ipv6_to_int(network.broadcast_address)
    network = _ipv6_to_int(network.network_address)
    return(network, broadcast)

def _address_range(ipv, address, width):
    "From an IP address create integers for the network and broadcast IP"
    # This is essentially the range which in between an IP address is.
    if ipv == 4:
        # IPv4, the width is given as the number of IPs
        network, broadcast = _address_range_ipv4(address, width)
    else:
        # IPv6, width is given by a netmask.
        network, broadcast = _address_range_ipv6(address, width)

    return (network, broadcast)

def _parse_row(row):
    "Parse and modify the row."
    columns = row.strip().split('|')
    # If there isn't more then 6 columns I can't parse it, so skipping it.
    if len(columns) > 6:
        tmp = columns[:5]

        if len(tmp[1].strip()) == 0:
            # This is the country it is assigned to, if there is no country
            # I am not interested in it.
            return None

        if tmp[2].strip().lower() not in ['ipv4', 'ipv6']:
            # If the protocol is not an IP protocol (such as asn), I am not
            # interested.
            return None

        if '6' in tmp[2]:
            tmp[2] = 6
        else:
            tmp[2] = 4

        # Convert the IP address and netmask/number of IP's to an IP range where
        # the IPs are converted to a numerical value.
        tmp[3], tmp[4] = _address_range(tmp[2], tmp[3], tmp[4])
        return tmp


class CompactRanges(object):
    "Try to compact the ranges."
    def __init__(self):
        self.ranges = list()

    def add(self, *newer):
        "Add a line to the ranges, compacting where possible."
        # nic, tld, ipv, network, broadcast = *newer
        newer = list(newer)

        if len(self.ranges) == 0:
            self.ranges.append(newer)
            return

        # Testing if current range is a continuation of the previous one
        older = self.ranges[-1]
        if older[0] == newer[0] and \
           older[1] == newer[1] and \
           older[2] == newer[2] and \
           older[4] == newer[3] - 1:
            # The older broadcast is the same as newer network - 1, thus is is a
            # continuation, so extending the range of the older one.
            self.ranges[-1][4] = newer[4]
        else:
            self.ranges.append(newer)

    def length(self):
        "return length of ranges"
        return len(self.ranges)


def _local_file_from_url(url):
    "Open the file, if available from the url"
    file_name = _split_url(url)[2]
    candidates = list()

    for candidate in os.listdir(TWD):
        if file_name.lower() in candidate.lower():
            candidates.append(candidate)

    candidates.sort(reverse=True)
    if len(candidates) == 0:
        print('# No files to parse')
        return None

    file_full = os.path.join(TWD, candidates[0])
    return file_full

def parse_latest(url):
    "Parse a file as it has been retrieved from the url."
    file_name = _local_file_from_url(url)
    if file_name is None:
        print('# No files available to parse !')
        return

    print('# Opening file: %s' % file_name)

    compacted = CompactRanges()
    count_linesall = 0
    count_relevant = 0
    with open(file_name, 'r') as file_open:
        for row in file_open:
            count_linesall += 1
            parsed = _parse_row(row)
            if parsed is None:
                continue

            count_relevant += 1
            compacted.add(*parsed)

    print('# Parsed %s lines' % count_linesall)
    print('#  - of which relevant: %s' % count_relevant)
    print('#  - reduced to ranges: %s' % compacted.length())
    return compacted.ranges

def _compact_string(text):
    "try making text compacter"
    # we go through the text and try to replace repeated characters with:
    # _c_n_  where c is the character and n is the amount of if. The underscore
    # in this context is guaranteed to not occur in text. As such we can use
    # it as an escape character.
    # Also we do not collapse if repeated character is below 5.
    tmp = list()
    last = ''
    count = 0
    for character in text+'_':
        # Add the underscore so we make sure not to miss the last bit of the
        # string if it happens to end on more then 4 identical characters.
        count += 1
        if character != last:
            if count > 4:
                tmp = tmp[:len(tmp)-count]
                tmp.append('_%s_%s_' % (last, count))

            count = 0

        last = character
        tmp.append(character)

    # Remove the appended underscore before returning.
    return ''.join(tmp)[:-1]

def get():
    "Fetch and parse data"
    print('#'*79)
    print('# Fetching data from regional NICs.')
    print('#'*79)
    tmp = list()
    for delegate in DELEGATES:
        print('# Using: %s' % delegate)
        download(delegate)
        tmp += parse_latest(delegate)
        print('#' * 79)
    print('# A total of %s IP ranges have been defined.' % len(tmp))

    for nic, country, ipv, network, broadcast in tmp:
        hex_network = hex(network)[2::].zfill(32)
        hex_broadcast = hex(broadcast)[2::].zfill(32)
        rid = nic[:2]+country+str(ipv)+hex_broadcast+hex_network
        rid = rid.lower()
        rid = _compact_string(rid)
        yield rid, nic, country, ipv, hex_network, hex_broadcast



"""
Fetch data.
"""
import ipaddress
from . import models

def country_by_ip(ip_address):
    "Return the TLD assigned to the address block the IP address is part of."
    address = ipaddress.ip_address(ip_address)
    hex_address = hex(int(address))[2::].zfill(32)

    model = models.ALL['IPRange']
    value = model.objects.filter(ipv=address.version,
                                 network_hex__lt=hex_address,
                                 broadcast_hex__gt=hex_address)

    count = value.count()
    if count < 1:
        return None
    elif count > 1:
        text = '! IPCountry table inconsistent on address %s' % ip_address
        raise ValueError(text)
    else:
        return value[0].tld


def currency_by_country_tld(tld):
    "Return the currency data by the top level domain."
    query = models.ALL['Currency'].objects.filter(reference__code_2=tld)
    count = query.count()

    if count >= 1:
        return query[0].code




